#include "StdAfx.h"
#include ".\container.h"
#include "RootContainer.h"
#include "Binding.h"

IMPLEMENT_RTTI(CContainer, CComponent)

CContainer::CContainer(void)
{
	m_pLayoutManager = NULL;
	ptScroll.x = 0;
	ptScroll.y = 0;	
	m_Padding.left = m_Padding.top = m_Padding.right = m_Padding.bottom = 0;
	m_DataContext = NULL;
}

CContainer::~CContainer(void)
{
	if (m_pLayoutManager != NULL)
	{
		m_pLayoutManager->AutoRelease();
		m_pLayoutManager = NULL;
	}
	
	if (m_DataContext != NULL)
	{
		m_DataContext->AutoRelease();
		m_DataContext = NULL;
	}
	
	RemoveAll();
}
void CContainer::SetDataContext(IDataContext* context)
{
	if (m_DataContext != NULL)
	{
		m_DataContext->AutoRelease();
		m_DataContext = NULL;
	}
	m_DataContext = context;
	if (m_DataContext != NULL)
	{
		m_DataContext->AddRef();
	}
}
IDataContext* CContainer::GetDataContext()
{
	return m_DataContext;
}
IDataContext* CContainer::FindDataContext(CContainer** ppFindContainer)
{
	IDataContext* datacontext = NULL;

	CContainer* pParent = this;
	while (pParent)
	{
		datacontext = pParent->GetDataContext();
		if (datacontext != NULL)
		{
			if (ppFindContainer != NULL)
			{
				*ppFindContainer = pParent;
			}			
			break;
		}
		pParent = pParent->GetParent();
	}
	return datacontext;
}
void CContainer::Add(CComponent* pChild)
{
	ATLASSERT(pChild);
	if (NULL != pChild)
	{	
		POSITION pos = NULL;
		pos = m_vComponent.Find(pChild, pos);
		if (pos)
		{
			return;
		}
		m_vComponent.AddTail(pChild);
		pChild->SetParent(this);
		pChild->OnAddNotify();
		this->OnAddChild(pChild);
	}
}
BOOL CContainer::Remove(CComponent* pChild)
{
	ATLASSERT(pChild);
	POSITION pos = NULL;
	pos = m_vComponent.Find(pChild, pos);
	if (pos)
	{
		this->OnRemoveChild(pChild);
		pChild->OnRemoveNotify();
		m_vComponent.RemoveAt(pos);
		pChild->SetParent(NULL);
		pChild->AutoRelease();
		return TRUE;
	}
	return FALSE;
}
BOOL CContainer::RemoveAll()
{
	POSITION pos = m_vComponent.GetHeadPosition();
	while (pos)
	{
		CComponent* pComponent = m_vComponent.GetNext(pos);
		if (pComponent)
		{
			pComponent->RemoveAll();
			Remove(pComponent);
		}
	}
	m_vComponent.RemoveAll();
	return TRUE;
}
void CContainer::SetLayoutManager(ILayoutManager* pLayoutManager)
{
	if (m_pLayoutManager != NULL)
	{
		m_pLayoutManager->AutoRelease();
		m_pLayoutManager = NULL;
	}
	m_pLayoutManager = pLayoutManager;
	if (m_pLayoutManager != NULL)
	{
		m_pLayoutManager->AddRef();
	}	
}
ILayoutManager* CContainer::GetLayoutManager()
{
	return m_pLayoutManager;
}
void CContainer::RecalcLayout(BOOL bRecursive)
{
	if (bRecursive)
	{
		POSITION pos = m_vComponent.GetHeadPosition();
		while (pos)
		{
			CComponent* pComponent = m_vComponent.GetNext(pos);
			if (pComponent)
			{
				pComponent->RecalcLayout(bRecursive);
			}
		}
	}
	CRect rcBound;
	GetBounds(&rcBound);
	OnSize(rcBound.Width(), rcBound.Height());
}
CComponent* CContainer::GetComponentByName(LPCTSTR pszcName)
{	
	POSITION pos = m_vComponent.GetHeadPosition();
	while (pos)
	{
		CComponent* pComponent = m_vComponent.GetNext(pos);
		if (pComponent)
		{
			if (_tcsicmp(pszcName, pComponent->GetName()) == 0)
			{
				return pComponent;
			}
		}
	}
	return NULL;
}
CComponent* CContainer::FindName(LPCTSTR pszcName)
{
	CComponent* pFindComponent = NULL;
	if (_tcsicmp(pszcName, this->GetName()) == 0)
	{
		pFindComponent = this;
		return pFindComponent;
	}
	POSITION pos = m_vComponent.GetHeadPosition();
	while (pos)
	{
		CComponent* pComponent = m_vComponent.GetNext(pos);
		if (pComponent)
		{
			if (_tcsicmp(pszcName, pComponent->GetName()) == 0)
			{
				pFindComponent = pComponent;
				break;				
			}
			CContainer* pContainer = dynamic_cast<CContainer*>(pComponent);
			if (pContainer != NULL)
			{
				CComponent* pTempComponent = pContainer->FindName(pszcName);
				if (pTempComponent != NULL)
				{
					pFindComponent = pTempComponent;
					break;
				}
			}
		}
	}
	return pFindComponent;
}
CComponent* CContainer::FindParentName(LPCTSTR pszcName)
{
	CComponent* pFindComponent = __super::FindParentName(pszcName);
	if (pFindComponent != NULL) return pFindComponent;

	CComponent* pParent = this;
	while (pParent != NULL)
	{
		pParent = pParent->GetParent();
		if (pParent == NULL)
			break;
		if (_tcsicmp(pszcName, pParent->GetName()) == 0)
		{
			pFindComponent = pParent;
			break;				
		}
	}
	return pFindComponent;
}
void CContainer::GetPadding(LPRECT padding)
{
	if (padding)
	{
		CopyRect(padding, &m_Padding);
	}
}
void CContainer::SetPadding(LPRECT padding)
{
	if (padding)
	{
		CopyRect(&m_Padding, padding);
	}
}
void CContainer::SetVisible(BOOL bVisible)
{
	__super::SetVisible(bVisible);
	
	POSITION pos = m_vComponent.GetHeadPosition();
	while (pos)
	{
		CComponent* pComponent = m_vComponent.GetNext(pos);
		if (pComponent)
		{
			pComponent->OnParentVisibleChanged(this, bVisible);
		}
	}
}
void CContainer::OnParentVisibleChanged(CComponent* parent, BOOL bVisible)
{
	POSITION pos = m_vComponent.GetHeadPosition();
	while (pos)
	{
		CComponent* pComponent = m_vComponent.GetNext(pos);
		if (pComponent)
		{
			pComponent->OnParentVisibleChanged(parent, bVisible);
		}
	}
}
void CContainer::OnDraw(HDC hDC)
{
	this->DrawComponent(hDC);
	this->DrawChild(hDC);	
}
void CContainer::OnSize(INT32 nWidth, INT32 nHeight)
{
	__super::OnSize(nWidth, nHeight);

	CAtlList<CComponent*> vAllComponent;
	_GetAllComponent(vAllComponent);

	POSITION pos = vAllComponent.GetHeadPosition();
	while (pos)
	{
		CComponent* pComponent = vAllComponent.GetNext(pos);
		if (pComponent)
		{
			pComponent->OnAncestorResized();
		}
	}
	if (m_pLayoutManager)
	{
		m_pLayoutManager->LayoutContainer(this);
	}
}
void CContainer::OnMouseLeave()
{
	if(this->_IsWindowAccessible() == FALSE) return;
	__super::OnMouseLeave();

	CAtlList<CComponent*> vAllComponent;
	_GetAllComponent(vAllComponent);

	POSITION pos = vAllComponent.GetTailPosition();
	while (pos)
	{
		CComponent* pComponent = vAllComponent.GetPrev(pos);
		if (pComponent)
		{
			pComponent->OnMouseLeave();
		}
	}

}
void CContainer::MouseAway()
{	
	this->OnMouseLeave();
}

void CContainer::DrawChild(HDC hDC)
{
	HRGN hRgnOldClip = CreateRectRgn(0, 0, 0, 0);
	int nResultGetClip = GetClipRgn(hDC, hRgnOldClip);
	
	CRect rcBounds;
	GetBounds(&rcBounds);
	POINT ptAbsolute;
	GetParentAbsolutePosition(&ptAbsolute);
	rcBounds.OffsetRect(ptAbsolute);
		
	POINT ptScroll = {0, 0};
	if (FALSE == _GetScrollPos(&ptScroll))
	{
		ptScroll.x = 0;
		ptScroll.y = 0;
	}
	HRGN hRgnClip = CreateRectRgn(rcBounds.left, rcBounds.top, rcBounds.right, rcBounds.bottom);

	if (nResultGetClip == 1)
	{
		CombineRgn(hRgnClip, hRgnClip, hRgnOldClip, RGN_AND);
	}
	SelectClipRgn(hDC, hRgnClip);
	
	POINT ptOrgChild = {0, 0};	
	BOOL bChangeViewportChild = FALSE;

	CRect rcChildBound;
	POSITION pos = m_vComponent.GetHeadPosition();
	while (pos)
	{
		CComponent* pComponent = m_vComponent.GetNext(pos);
		if (pComponent)
		{
			if (pComponent->IsVisible())
			{
				pComponent->GetBounds(&rcChildBound);
				bChangeViewportChild = OffsetViewportOrgEx(hDC, rcChildBound.left - ptScroll.x, 
															rcChildBound.top - ptScroll.y, &ptOrgChild);
				
				pComponent->OnDraw(hDC);
				if (bChangeViewportChild)
				{
					SetViewportOrgEx(hDC, ptOrgChild.x, ptOrgChild.y, NULL);
				}
			}			
		}
	}

	switch (nResultGetClip)
	{
	case 0:
		SelectClipRgn(hDC, NULL);
		break;
	case 1:
		SelectClipRgn(hDC, hRgnOldClip);
		break;
	default:
		break;
	}
	DeleteObject(hRgnClip);
	DeleteObject(hRgnOldClip);
	
	this->DrawNCChild(hDC);
}
void CContainer::DrawNCChild(HDC hDC)
{
	HRGN hRgnOldClip = CreateRectRgn(0, 0, 0, 0);
	int nResultGetClip = GetClipRgn(hDC, hRgnOldClip);
	//ATLASSERT(nResultGetClip != -1);
	CRect rcBound;
	GetBounds(&rcBound);
	CRect rcBounds;
	GetBounds(&rcBounds);
	POINT ptAbsolute;
	GetParentAbsolutePosition(&ptAbsolute);
	rcBounds.OffsetRect(ptAbsolute);
	
	HRGN hRgnClip = CreateRectRgn(rcBounds.left, rcBounds.top, rcBounds.right, rcBounds.bottom);
	if (nResultGetClip == 1)
	{
		CombineRgn(hRgnClip, hRgnClip, hRgnOldClip, RGN_AND);
	}
	SelectClipRgn(hDC, hRgnClip);

	POINT ptOrg = {0, 0};	
	BOOL bChangeViewport = FALSE;

	CRect rcChildBound;
	POSITION pos = m_vNCComponent.GetHeadPosition();
	while (pos)
	{
		CComponent* pComponent = m_vNCComponent.GetNext(pos);
		if (pComponent)
		{
			if (pComponent->IsVisible())
			{
				pComponent->GetBounds(&rcChildBound);
				bChangeViewport = OffsetViewportOrgEx(hDC, rcChildBound.left, rcChildBound.top, &ptOrg);

				pComponent->OnDraw(hDC);
				if (bChangeViewport)
				{
					SetViewportOrgEx(hDC, ptOrg.x, ptOrg.y, NULL);
				}
			}			
		}
	}

	switch (nResultGetClip)
	{
	case 0:
		SelectClipRgn(hDC, NULL);
		break;
	case 1:
		SelectClipRgn(hDC, hRgnOldClip);
		break;
	default:
		break;
	}
	DeleteObject(hRgnClip);
	DeleteObject(hRgnOldClip);
}
void CContainer::AddNCComponent(CComponent* pChild)
{
	ATLASSERT(pChild);
	if (NULL != pChild)
	{	
		m_vNCComponent.AddTail(pChild);
		pChild->SetParent(this);
		pChild->OnAddNotify();		
	}
}
BOOL CContainer::RemoveNCComponent(CComponent* pChild)
{
	ATLASSERT(pChild);
	POSITION pos = NULL;
	pos = m_vNCComponent.Find(pChild, pos);
	if (pos)
	{
		pChild->OnRemoveNotify();
		m_vNCComponent.RemoveAt(pos);
		pChild->SetParent(NULL);
		return TRUE;
	}
	return FALSE;
}
BOOL CContainer::ContainedNCComponent(CComponent* pComponent)
{
	POSITION pos = NULL;
	pos = m_vNCComponent.Find(pComponent, pos);
	if (pos)
	{
		return TRUE;
	}
	return FALSE;
}
BOOL CContainer::_GetAllComponent(CAtlList<CComponent*>& vComponent)
{
	vComponent.AddTailList(&m_vComponent);
	vComponent.AddTailList(&m_vNCComponent);
	return TRUE;
}
BOOL CContainer::GetPreferredSize(LPSIZE lpSize)
{
	BOOL bSuccess = FALSE;
	CSize sizePreferred(0, 0);
	bSuccess = __super::GetPreferredSize(&sizePreferred);

	if (sizePreferred.cx == 0 || sizePreferred.cy == 0)
	{
		CSize sizePreferredChild(0, 0);
		if (this->GetPreferredSizeFromChild(&sizePreferredChild))
		{
			if (sizePreferred.cx == 0)
			{
				sizePreferred.cx = sizePreferredChild.cx;
			}
			if (sizePreferred.cy == 0)
			{
				sizePreferred.cy = sizePreferredChild.cy;
			}
			bSuccess = TRUE;
		}
	}
	lpSize->cx = sizePreferred.cx;
	lpSize->cy = sizePreferred.cy;	
	return bSuccess;
}
BOOL CContainer::GetPreferredSizeFromChild(LPSIZE lpSize)
{
	BOOL bSuccess = FALSE;
	CSize sizePreferred(0, 0);
	CAtlList<CComponent*> vAllComponent;
	_GetAllComponent(vAllComponent);
	if (vAllComponent.GetCount() > 0)
	{
		CRect rcComponentBound;
		CSize sizeComponentPreferred;
		POSITION pos = vAllComponent.GetHeadPosition();
		while (pos)
		{
			CComponent* pComponent = vAllComponent.GetNext(pos);
			if (pComponent)
			{
				if (pComponent->IsVisible() && TRUE == pComponent->GetPreferredSize(&sizeComponentPreferred))
				{
					pComponent->GetBounds(&rcComponentBound);
					sizePreferred.cx = max(sizePreferred.cx, rcComponentBound.left + sizeComponentPreferred.cx);
					sizePreferred.cy = max(sizePreferred.cy, rcComponentBound.top + sizeComponentPreferred.cy);					
				}				
			}
		}
		bSuccess = TRUE;
	}
	if (bSuccess)
	{
		lpSize->cx = sizePreferred.cx;
		lpSize->cy = sizePreferred.cy;
	}
	return bSuccess;
}

BOOL CContainer::PreTranslateMessage(MSG* pMsg)
{
	CAtlList<CComponent*> vAllComponent;
	_GetAllComponent(vAllComponent);
	POSITION pos = vAllComponent.GetHeadPosition();
	while (pos)
	{
		CComponent* pComponent = vAllComponent.GetNext(pos);
		if (pComponent)
		{
			if( pComponent->PreTranslateMessage(pMsg) == TRUE )
				return TRUE;
		}
	}

	return FALSE;
}
CComponent* CContainer::_GetCapturedComponent()
{
	/// {kr} ĸĵ Ʈ, ĸĵ Ʈ ִ . {/kr}@n
	CComponent* pCapturedComponent = NULL;
	CRootContainer* pRootContainer = this->GetRootContainer();
	if (pRootContainer != NULL)
	{
		pCapturedComponent = pRootContainer->GetCapturedComponent();
	}
	return pCapturedComponent;
}